"""Kohn-Sham single particle excitations realated objects.

"""
import sys
import json
import numpy as np

from ase.units import Bohr, Hartree, alpha

import gpaw.mpi as mpi
from gpaw.utilities import packed_index
from gpaw.lrtddft.excitation import Excitation, ExcitationList, get_filehandle
from gpaw.pair_density import PairDensity
from gpaw.fd_operators import Gradient
from gpaw.utilities.tools import coordinates


class KSSingles(ExcitationList):
    """Kohn-Sham single particle excitations

    Input parameters:

    calculator:
      the calculator object after a ground state calculation

    nspins:
      number of spins considered in the calculation
      Note: Valid only for unpolarised ground state calculation

    eps:
      Minimal occupation difference for a transition (default 0.001)

    istart:
      First occupied state to consider
    jend:
      Last unoccupied state to consider
    energy_range:
      The energy range [emin, emax] or emax for KS transitions to use as basis
    """
    def __init__(self,
                 restrict={},
                 log=None,
                 txt=None):
        ExcitationList.__init__(self, log=log, txt=txt)
        self.world = mpi.world

        self.restrict = KSSRestrictor()
        self.restrict.update(restrict)

    def calculate(self, atoms, nspins=None):
        calculator = atoms.calc
        self.calculator = calculator

        # LCAO calculation requires special actions
        self.lcao = calculator.parameters.mode == 'lcao'

        # deny hybrids as their empty states are wrong
#        gsxc = calculator.hamiltonian.xc
#        hybrid = hasattr(gsxc, 'hybrid') and gsxc.hybrid > 0.0
#        assert(not hybrid)

        # ensure correctly initialized wave functions
        calculator.converge_wave_functions()
        self.world = calculator.wfs.world

        # parallelization over bands not yet supported
        assert(calculator.wfs.bd.comm.size == 1)

        # do the evaluation
        self.select(nspins)

        trkm = self.get_trk()
        self.log('KSS {0} transitions (restrict={1})'.format(
            len(self), self.restrict))
        self.log('KSS TRK sum %g (%g,%g,%g)' %
                 (np.sum(trkm) / 3., trkm[0], trkm[1], trkm[2]))
        pol = self.get_polarizabilities(lmax=3)
        self.log('KSS polarisabilities(l=0-3) %g, %g, %g, %g' %
                 tuple(pol.tolist()))
        return self

    @staticmethod
    def emin_emax(energy_range):
        emin = -sys.float_info.max
        emax = sys.float_info.max
        if energy_range is not None:
            try:
                emin, emax = energy_range
                emin /= Hartree
                emax /= Hartree
            except TypeError:
                emax = energy_range / Hartree
        return emin, emax

    def select(self, nspins=None):
        """Select KSSingles according to the given criterium."""

        # criteria
        emin, emax = self.restrict.emin_emax()
        istart = self.restrict['istart']
        jend = self.restrict['jend']
        eps = self.restrict['eps']

        if not hasattr(self, 'calculator'):  # I'm read from a file
            # throw away all not needed entries
            for i, ks in reversed(list(enumerate(self))):
                if ((ks.fij / ks.weight) <= eps or
                    ks.i < istart or ks.j > jend or
                    ks.energy < emin or ks.energy > emax):
                    del(self[i])
            return None

        paw = self.calculator
        wfs = paw.wfs
        self.dtype = wfs.dtype
        self.kpt_u = wfs.kpt_u

        if not self.lcao and self.kpt_u[0].psit_nG is None:
            raise RuntimeError('No wave functions in calculator!')

        # here, we need to take care of the spins also for
        # closed shell systems (Sz=0)
        # vspin is the virtual spin of the wave functions,
        #       i.e. the spin used in the ground state calculation
        # pspin is the physical spin of the wave functions
        #       i.e. the spin of the excited states
        self.nvspins = wfs.nspins
        self.npspins = wfs.nspins
        fijscale = 1
        ispins = [0]
        nks = wfs.kd.nibzkpts * wfs.kd.nspins
        if self.nvspins < 2:
            if (nspins or 0) > self.nvspins:
                self.npspins = nspins
                fijscale = 0.5
                ispins = [0, 1]
                nks *= 2

        kpt_comm = self.calculator.wfs.kd.comm
        nbands = len(self.kpt_u[0].f_n)

        # select
        take = np.zeros((nks, nbands, nbands), dtype=int)
        u = 0
        for ispin in ispins:
            for k in range(wfs.kd.nibzkpts):
                q = k - wfs.kd.k0
                for s in range(wfs.nspins):
                    if q >= 0 and q < wfs.kd.mynk:
                        kpt = wfs.kpt_qs[q][s]
                        for i in range(nbands):
                            for j in range(i + 1, nbands):
                                fij = (kpt.f_n[i] - kpt.f_n[j]) / kpt.weight
                                epsij = kpt.eps_n[j] - kpt.eps_n[i]
                                if (fij > eps and
                                    epsij >= emin and epsij < emax and
                                        i >= istart and j <= jend):
                                    take[u, i, j] = 1
                    u += 1
        kpt_comm.sum(take)

        self.log()
        self.log('Kohn-Sham single transitions')
        self.log()

        # calculate in parallel
        u = 0
        for ispin in ispins:
            for k in range(wfs.kd.nibzkpts):
                q = k - wfs.kd.k0
                for s in range(wfs.kd.nspins):
                    for i in range(nbands):
                        for j in range(i + 1, nbands):
                            if take[u, i, j]:
                                if q >= 0 and q < wfs.kd.mynk:
                                    kpt = wfs.kpt_qs[q][s]
                                    pspin = max(kpt.s, ispin)
                                    self.append(
                                        KSSingle(i, j, pspin, kpt, paw,
                                                 fijscale=fijscale,
                                                 dtype=self.dtype))
                                else:
                                    self.append(KSSingle(i, j, pspin=0,
                                                         kpt=None, paw=paw,
                                                         dtype=self.dtype))
                    u += 1

        # distribute
        for kss in self:
            kss.distribute()

    @classmethod
    def read(cls, filename=None, fh=None, restrict={}, log=None):
        """Read myself from a file"""
        assert (filename is not None) or (fh is not None)
        
        def fail(f):
            raise RuntimeError(f.name + ' does not contain ' +
                               cls.__class__.__name__ + ' data')
        if fh is None:
            f = get_filehandle(cls, filename)

            # there can be other information, i.e. the LrTDDFT header
            try:
                content = f.read()
                f.seek(content.index('# KSSingles'))
                del(content)
                f.readline()
            except ValueError:
                fail(f)
        else:
            f = fh
            # we assume to be at the right place and read the header
            if not f.readline().strip() == '# KSSingles':
                fail(f)

        words = f.readline().split()
        n = int(words[0])
        kssl = cls(log=log)
        if len(words) == 1:
            # very old output style for real wave functions (finite systems)
            kssl.dtype = float
            restrict_from_file = {}
        else:
            if words[1].startswith('complex'):
                kssl.dtype = complex
            else:
                kssl.dtype = float
            restrict_from_file = json.loads(f.readline())
            if not isinstance(restrict_from_file, dict):  # old output style
                restrict_from_file = {'eps': restrict_from_file}
        kssl.npspins = 1
        for i in range(n):
            kss = KSSingle(string=f.readline(), dtype=kssl.dtype)
            kssl.append(kss)
            kssl.npspins = max(kssl.npspins, kss.pspin + 1)
        
        if fh is None:
            f.close()

        kssl.update()
        kssl.restrict.update(restrict_from_file)
        if len(restrict):
            kssl.restrict.update(restrict)
            kssl.select()
        
        return kssl

    def update(self):
        istart = self[0].i
        jend = 0
        npspins = 1
        nvspins = 1
        for kss in self:
            istart = min(kss.i, istart)
            jend = max(kss.j, jend)
            if kss.pspin == 1:
                npspins = 2
            if kss.spin == 1:
                nvspins = 2
        self.restrict.update({'istart': istart, 'jend': jend})
        self.npspins = npspins
        self.nvspins = nvspins

        if hasattr(self, 'energies'):
            del(self.energies)

    def set_arrays(self):
        if hasattr(self, 'energies'):
            return
        energies = []
        fij = []
        me = []
        mur = []
        muv = []
        magn = []
        for k in self:
            energies.append(k.energy)
            fij.append(k.fij)
            me.append(k.me)
            mur.append(k.mur)
            if k.muv is not None:
                muv.append(k.muv)
            if k.magn is not None:
                magn.append(k.magn)
        self.energies = np.array(energies)
        self.fij = np.array(fij)
        self.me = np.array(me)
        self.mur = np.array(mur)
        if len(muv):
            self.muv = np.array(muv)
        else:
            self.muv = None
        if len(magn):
            self.magn = np.array(magn)
        else:
            self.magn = None

    def write(self, filename=None, fh=None):
        """Write current state to a file.

        'filename' is the filename. If the filename ends in .gz,
        the file is automatically saved in compressed gzip format.

        'fh' is a filehandle. This can be used to write into already
        opened files.
        """
        if self.world.rank != 0:
            return

        if fh is None:
            f = get_filehandle(self, filename, mode='w')
        else:
            f = fh

        f.write('# KSSingles\n')
        f.write('{0} {1}\n'.format(len(self), np.dtype(self.dtype)))
        f.write(json.dumps(self.restrict.values) + '\n')
        for kss in self:
            f.write(kss.outstring())
        if fh is None:
            f.close()

    def overlap(self, ov_nn, other):
        """Matrix element overlaps determined from wave function overlaps.

        Parameters
        ----------
        ov_nn: array
            Wave function overlap factors from a displaced calculator.
            Index 0 corresponds to our own wavefunctions conjugated and
            index 1 to the others' wavefunctions

        Returns
        -------
        ov_pp: array
            Overlap corresponding to matrix elements.
            Index 0 corresponds to our own matrix elements conjugated and
            index 1 to the others' matrix elements
        """
        n0 = len(self)
        n1 = len(other)
        ov_pp = np.zeros((n0, n1), dtype=ov_nn.dtype)
        i1_p = [ex.i for ex in other]
        j1_p = [ex.j for ex in other]
        for p0, ex0 in enumerate(self):
            ov_pp[p0, :] = ov_nn[ex0.i, i1_p].conj() * ov_nn[ex0.j, j1_p]
        return ov_pp


class KSSRestrictor:
    """Object to handle KSSingles restrictions"""
    defaults = {'eps': 0.01,
                'istart': 0,
                'jend': sys.maxsize,
                'energy_range': None}
    
    def __init__(self, dictionary={}):
        self._vals = {}
        self.update(dictionary)

    def __getitem__(self, index):
        assert index in self.defaults
        return self._vals.get(index, self.defaults[index])
    
    def __setitem__(self, index, value):
        assert index in self.defaults
        self._vals[index] = value
        
    def update(self, dictionary):
        for key, value in dictionary.items():
            self[key] = value
            
    def emin_emax(self):
        emin = -sys.float_info.max
        emax = sys.float_info.max
        if self['energy_range'] is not None:
            try:
                emin, emax = self['energy_range']
                emin /= Hartree
                emax /= Hartree
            except TypeError:
                emax = self['energy_range'] / Hartree
        return emin, emax

    def __ge__(self, other):
        """am I less or equal restricting than the other?"""
        emin, emax = self.emin_emax()
        omin, omax = other.emin_emax()
        res = (emin <= omin) & (emax >= omax)
        
        res &= self['istart'] <= other['istart']
        res &= self['jend'] >= other['jend']
        res &= self['eps'] < other['eps']

        return res

    @property
    def values(self):
        return dict(self._vals)

    def __str__(self):
        return str(self.values)


class KSSingle(Excitation, PairDensity):
    """Single Kohn-Sham transition containing all it's indicees

      pspin=physical spin
      spin=virtual  spin, i.e. spin in the ground state calc.
      kpt=the Kpoint object
      fijscale=weight for the occupation difference::
      me  = sqrt(fij*epsij) * <i|r|j>
      mur = - <i|r|a>
      muv = - <i|nabla|a>/omega_ia with omega_ia>0
      magn = <i|[r x nabla]|a> / (2 m_e c)
    """

    def __init__(self, iidx=None, jidx=None, pspin=None, kpt=None,
                 paw=None, string=None, fijscale=1, dtype=float):

        if string is not None:
            self.fromstring(string, dtype)
            return None

        # normal entry

        PairDensity.__init__(self, paw)
        PairDensity.initialize(self, kpt, iidx, jidx)

        self.pspin = pspin

        self.energy = 0.0
        self.fij = 0.0

        self.me = np.zeros((3), dtype=dtype)
        self.mur = np.zeros((3), dtype=dtype)
        self.muv = np.zeros((3), dtype=dtype)
        self.magn = np.zeros((3), dtype=dtype)

        self.kpt_comm = paw.wfs.kd.comm

        # leave empty if not my kpt
        if kpt is None:
            return

        wfs = paw.wfs
        gd = wfs.gd

        self.energy = kpt.eps_n[jidx] - kpt.eps_n[iidx]
        self.fij = (kpt.f_n[iidx] - kpt.f_n[jidx]) * fijscale

        # calculate matrix elements -----------

        # length form ..........................

        # course grid contribution
        # <i|r|j> is the negative of the dipole moment (because of negative
        # e- charge)
        me = - gd.calculate_dipole_moment(self.get())

        # augmentation contributions
        ma = np.zeros(me.shape, dtype=dtype)
        pos_av = paw.atoms.get_positions() / Bohr
        for a, P_ni in kpt.P_ani.items():
            Ra = pos_av[a]
            Pi_i = P_ni[self.i].conj()
            Pj_i = P_ni[self.j]
            Delta_pL = wfs.setups[a].Delta_pL
            ni = len(Pi_i)
            ma0 = 0
            ma1 = np.zeros(me.shape, dtype=me.dtype)
            for i in range(ni):
                for j in range(ni):
                    pij = Pi_i[i] * Pj_i[j]
                    ij = packed_index(i, j, ni)
                    # L=0 term
                    ma0 += Delta_pL[ij, 0] * pij
                    # L=1 terms
                    if wfs.setups[a].lmax >= 1:
                        # see spherical_harmonics.py for
                        # L=1:y L=2:z; L=3:x
                        ma1 += np.array([Delta_pL[ij, 3], Delta_pL[ij, 1],
                                         Delta_pL[ij, 2]]) * pij
            ma += np.sqrt(4 * np.pi / 3) * ma1 + Ra * np.sqrt(4 * np.pi) * ma0
        gd.comm.sum(ma)

        self.me = np.sqrt(self.energy * self.fij) * (me + ma)
        self.mur = - (me + ma)

        # velocity form .............................

        if self.lcao:
            # XXX Velocity form not supported in LCAO
            return

        me = np.zeros(self.mur.shape, dtype=dtype)

        # get derivatives
        dtype = self.wfj.dtype
        dwfj_cg = gd.empty((3), dtype=dtype)
        if not hasattr(gd, 'ddr'):
            gd.ddr = [Gradient(gd, c, dtype=dtype).apply for c in range(3)]
        for c in range(3):
            gd.ddr[c](self.wfj, dwfj_cg[c], kpt.phase_cd)
            me[c] = gd.integrate(self.wfi.conj() * dwfj_cg[c])

        if 0:
            me2 = np.zeros(self.mur.shape)
            for c in range(3):
                gd.ddr[c](self.wfi, dwfj_cg[c], kpt.phase_cd)
                me2[c] = gd.integrate(self.wfj * dwfj_cg[c])
            print(me, -me2, me2 + me)

        # augmentation contributions
        ma = np.zeros(me.shape, dtype=me.dtype)
        for a, P_ni in kpt.P_ani.items():
            Pi_i = P_ni[self.i].conj()
            Pj_i = P_ni[self.j]
            nabla_iiv = paw.wfs.setups[a].nabla_iiv
            for c in range(3):
                for i1, Pi in enumerate(Pi_i):
                    for i2, Pj in enumerate(Pj_i):
                        ma[c] += Pi * Pj * nabla_iiv[i1, i2, c]
        gd.comm.sum(ma)

        self.muv = - (me + ma) / self.energy

        # magnetic transition dipole ................

        r_cg, r2_g = coordinates(gd)
        magn = np.zeros(me.shape, dtype=dtype)

        wfi_g = self.wfi.conj()
        for ci in range(3):
            cj = (ci + 1) % 3
            ck = (ci + 2) % 3
            magn[ci] = gd.integrate(wfi_g * r_cg[cj] * dwfj_cg[ck] -
                                    wfi_g * r_cg[ck] * dwfj_cg[cj])
        # augmentation contributions
        ma = np.zeros(magn.shape, dtype=magn.dtype)
        for a, P_ni in kpt.P_ani.items():
            Pi_i = P_ni[self.i].conj()
            Pj_i = P_ni[self.j]
            rnabla_iiv = paw.wfs.setups[a].rnabla_iiv
            for c in range(3):
                for i1, Pi in enumerate(Pi_i):
                    for i2, Pj in enumerate(Pj_i):
                        ma[c] += Pi * Pj * rnabla_iiv[i1, i2, c]
        gd.comm.sum(ma)

        self.magn = -alpha / 2. * (magn + ma)

    def distribute(self):
        """Distribute results to all cores."""
        self.spin = self.kpt_comm.sum(self.spin)
        self.pspin = self.kpt_comm.sum(self.pspin)
        self.k = self.kpt_comm.sum(self.k)
        self.weight = self.kpt_comm.sum(self.weight)
        self.energy = self.kpt_comm.sum(self.energy)
        self.fij = self.kpt_comm.sum(self.fij)

        self.kpt_comm.sum(self.me)
        self.kpt_comm.sum(self.mur)
        self.kpt_comm.sum(self.muv)
        self.kpt_comm.sum(self.magn)

    def __add__(self, other):
        """Add two KSSingles"""
        result = self.copy()
        result.me = self.me + other.me
        result.mur = self.mur + other.mur
        result.muv = self.muv + other.muv
        result.magn = self.magn - other.magn
        return result

    def __sub__(self, other):
        """Subtract two KSSingles"""
        result = self.copy()
        result.me = self.me - other.me
        result.mur = self.mur - other.mur
        result.muv = self.muv - other.muv
        result.magn = self.magn - other.magn
        return result

    def __rmul__(self, x):
        return self.__mul__(x)

    def __mul__(self, x):
        """Multiply a KSSingle with a number"""
        if isinstance(x, (float, int)):
            result = self.copy()
            result.me = self.me * x
            result.mur = self.mur * x
            result.muv = self.muv * x
            return result
        else:
            return RuntimeError('not a number')

    def __truediv__(self, x):
        return self.__mul__(1. / x)

    __div__ = __truediv__

    def copy(self):
        if self.mur.dtype == complex:
            return KSSingle(string=self.outstring(), dtype=complex)
        else:
            return KSSingle(string=self.outstring(), dtype=float)

    def fromstring(self, string, dtype=float):
        l = string.split()
        self.i = int(l.pop(0))
        self.j = int(l.pop(0))
        self.pspin = int(l.pop(0))
        self.spin = int(l.pop(0))
        if dtype == float:
            self.k = 0
            self.weight = 1
        else:
            self.k = int(l.pop(0))
            self.weight = float(l.pop(0))
        self.energy = float(l.pop(0))
        self.fij = float(l.pop(0))
        self.mur = np.array([dtype(l.pop(0)) for i in range(3)])
        self.me = - self.mur * np.sqrt(self.energy * self.fij)
        self.muv = self.magn = None
        if len(l):
            self.muv = np.array([dtype(l.pop(0)) for i in range(3)])
        if len(l):
            self.magn = np.array([dtype(l.pop(0)) for i in range(3)])
        return None

    def outstring(self):
        if self.mur.dtype == float:
            string = '{0:d} {1:d}  {2:d} {3:d}  {4:g} {5:g}'.format(
                self.i, self.j, self.pspin, self.spin, self.energy, self.fij)
        else:
            string = (
                '{0:d} {1:d}  {2:d} {3:d} {4:d} {5:g}  {6:g} {7:g}'.format(
                    self.i, self.j, self.pspin, self.spin, self.k,
                    self.weight, self.energy, self.fij))
        string += '  '

        def format_me(me):
            string = ''
            if me.dtype == float:
                for m in me:
                    string += ' {0:.5e}'.format(m)
            else:
                for m in me:
                    string += ' {0.real:.5e}{0.imag:+.5e}j'.format(m)
            return string

        string += '  ' + format_me(self.mur)
        if self.muv is not None:
            string += '  ' + format_me(self.muv)
        if self.magn is not None:
            string += '  ' + format_me(self.magn)
        string += '\n'

        return string

    def __str__(self):
        string = '# <KSSingle> %d->%d %d(%d) eji=%g[eV]' % \
            (self.i, self.j, self.pspin, self.spin,
             self.energy * Hartree)
        if self.me.dtype == float:
            string += ' (%g,%g,%g)' % (self.me[0], self.me[1], self.me[2])
        else:
            string += ' kpt={0:d} w={1:g}'.format(self.k, self.weight)
            string += ' ('
            # use velocity form
            s = - np.sqrt(self.energy * self.fij)
            for c, m in enumerate(s * self.me):
                string += '{0.real:.5e}{0.imag:+.5e}j'.format(m)
                if c < 2:
                    string += ','
            string += ')'
        return string

    def __eq__(self, other):
        """KSSingles are considred equal when their indices are equal."""
        return (self.pspin == other.pspin and self.k == other.k and
                self.i == other.i and self.j == other.j)

    def __hash__(self):
        """Hash similar to __eq__"""
        if not hasattr(self, 'hash'):
            self.hash = hash((self.spin, self.k, self.i, self.j))
        return self.hash

    #
    # User interface: ##
    #

    def get_weight(self):
        return self.fij



import numpy as np
from math import pi, cos, sin
from ase import Atom, Atoms
from gpaw import GPAW, setup_paths
from gpaw.poisson import FDPoissonSolver
from gpaw.xas import XAS
from gpaw.test import equal
from gpaw.atom.generator2 import generate
import gpaw.mpi as mpi


def test_corehole_h2o(in_tmp_dir):
    # Generate setup for oxygen with half a core-hole:
    gen = generate('O', 8, '2s,s,2p,p,d', [1.2], 1.0, None, 2,
                   core_hole='1s,0.5')
    setup = gen.make_paw_setup('hch1s')
    setup.write_xml()
    if setup_paths[0] != '.':
        setup_paths.insert(0, '.')

    a = 5.0
    d = 0.9575
    t = pi / 180 * 104.51
    H2O = Atoms([Atom('O', (0, 0, 0)),
                 Atom('H', (d, 0, 0)),
                 Atom('H', (d * cos(t), d * sin(t), 0))],
                cell=(a, a, a), pbc=False)
    H2O.center()
    calc = GPAW(nbands=10, h=0.2, setups={'O': 'hch1s'},
                experimental={'niter_fixdensity': 2},
                poissonsolver=FDPoissonSolver(use_charge_center=True))
    H2O.calc = calc
    _ = H2O.get_potential_energy()

    if mpi.size == 1:
        xas = XAS(calc)
        x, y = xas.get_spectra()
        e1_n = xas.eps_n
        de1 = e1_n[1] - e1_n[0]

    calc.write('h2o-xas.gpw')

    if mpi.size == 1:
        calc = GPAW('h2o-xas.gpw', txt=None,
                    poissonsolver=FDPoissonSolver(use_charge_center=True))
        calc.initialize()
        xas = XAS(calc)
        x, y = xas.get_spectra()
        e2_n = xas.eps_n
        w_n = np.sum(xas.sigma_cn.real**2, axis=0)
        de2 = e2_n[1] - e2_n[0]

        equal(de2, 2.064, 0.005)
        equal(w_n[1] / w_n[0], 2.22, 0.01)

        assert de1 == de2

    if 0:
        import matplotlib.pyplot as plt
        plt.plot(x, y[0])
        plt.plot(x, sum(y))
        plt.show()



from ase.io.xyz import read_xyz

from gpaw import GPAW, Mixer, ConvergenceError, RMMDIIS, setup_paths

# Use setups from the $PWD and $PWD/.. first
setup_paths.insert(0, '.')
setup_paths.insert(0, '../')

atoms = read_xyz('../Au102_revised.xyz')

prefix = 'Au_cluster'
L = 32.0
atoms.set_cell((L, L, L), scale_atoms=False)
atoms.center()
atoms.set_pbc(1)
r = [1, 1, 1]
atoms = atoms.repeat(r)
n = [240 * ri for ri in r]
# nbands (>=1683) is the number of bands per cluster
nbands = 3 * 6 * 6 * 16  # 1728
for ri in r:
    nbands = nbands * ri
mixer = Mixer(beta=0.1, nmaxold=5, weight=100.0)
# the next three lines decrease memory usage
es = RMMDIIS(keep_htpsit=False)
calc = GPAW(nbands=nbands,
            # uncomment next two lines to use lcao/sz
            # mode='lcao',
            # basis='sz',
            gpts=tuple(n),
            maxiter=5,
            width=0.1,
            xc='LDA',
            mixer=mixer,
            eigensolver=es,
            txt=prefix + '.txt')

atoms.calc = calc
try:
    pot = atoms.get_potential_energy()
except ConvergenceError:
    pass



from myqueue.task import task


def create_tasks():
    return [task('gs.py@8:10m'),
            task('td.py@8:30m', deps='gs.py'),
            task('tdc.py@8:30m', deps='td.py'),
            task('td_replay.py@8:30m', deps='tdc.py'),
            task('spectrum.py@1:2m', deps='tdc.py'),
            task('td_fdm_replay.py@1:5m', deps='tdc.py'),
            task('ksd_init.py@1:5m', deps='gs.py'),
            task('fdm_ind.py@1:2m', deps='td_fdm_replay.py,ksd_init.py'),
            task('spec_plot.py@1:2m', deps='spectrum.py'),
            task('tcm_plot.py@1:2m',
                 deps='ksd_init.py,td_fdm_replay.py,spectrum.py'),
            task('ind_plot.py@1:2m', deps='fdm_ind.py'),
            task('td_pulse.py@8:30m', deps='gs.py'),
            task('plot_pulse.py@1:2m', deps='td_pulse.py'),
            ]



from gpaw import GPAW

print('state    LDA        PBE')

for name in ['ferro', 'anti', 'non']:

    calc = GPAW(name + '.gpw', txt=None)
    atoms = calc.get_atoms()
    eLDA = atoms.get_potential_energy()
    deltaxc = calc.get_xc_difference('PBE')
    ePBE = eLDA + deltaxc

    if name == 'ferro':
        eLDA0 = eLDA
        ePBE0 = ePBE

    eLDA -= eLDA0
    ePBE -= ePBE0
    print(('%-5s: %7.3f eV %7.3f eV' % (name, eLDA, ePBE)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Emacs: treat this as -*- python -*-
# flake8: noqa
from optparse import OptionParser

parser = OptionParser(usage='%prog [options] output_prefix.\nExample of call:\n'+
                      'python %prog --dir=. --pattern="b256H2O_120_04x04m64.grid_*_" b256H2O\n',
                      version='%prog 0.1')
parser.add_option('--dir', dest="dir",
                  default='.',
                  help='results directory.')
parser.add_option('--iter', dest="iter",
                  default=5,
                  help='number of SCF steps.')
parser.add_option('--pattern', dest="pattern",
                  default='',
                  help='pattern for directories to search.')
parser.add_option('-v', '--verbose', action='store_true',
                  default=False,
                  help='verbose mode.')

opt, args = parser.parse_args()

import os

def T(s):
    """Return time in seconds from hh:mm:ss"""
    t = 0
    for x in s.split(':'):
        t = t * 60 + int(x)
    return t

def analyse_benchmark(dir, pattern, output_prefix, iter, verbose=False):

    from os.path import abspath, exists, join

    output = output_prefix+'.txt'

    root_abspath = abspath(dir)
    # length of directory name
    rootlen = len(root_abspath) + 1

    assert exists(root_abspath)

    from glob import glob

    f_list = sorted(glob(join(root_abspath, pattern, output)))
    assert not f_list == [], 'Error: list of output files is empty'

    import re

    time = {}
    processes = []

    for f in f_list:
        # extract the number of processes p
        d = f.split('/')[-2]
        p = d.split('_')[-2]
        p = int(p)
        processes.append(p)
        #
        time[p] = {
        'start': 0.0,
        'fixdensity_start_estimate': 0.0,
        'fixdensity_end': 0.0,
        'SCF_start': 0.0,
        'SCF_end': 0.0,
        'forces_start': 0.0,
        'forces_end': 0.0,
        'total': 0.0,
        }
        #
        lines = open(f).readlines()
        # extract gpaw version
        for n, l in enumerate(lines):
            if l.startswith(' |__ |  _|___|_____|'):
                gpaw_version = lines[n + 0].strip().split()[3].split('.')[-1]
                break
        if gpaw_version[-1] == 'M':
            gpaw_version = gpaw_version[:-1]
        if gpaw_version.rfind(':') != -1:
            gpaw_version = gpaw_version[:gpaw_version.rfind(':')]
        gpaw_version = int(gpaw_version)
        # assume old version (< 3172)
        start_iter = 0
        # old style
        #
        #            Atomic orbitals used for initialization: 1536
        #                     log10-error:    Total        Iterations:
        #           Time      WFS    Density  Energy       Fermi  Poisson
        #iter:   0  12:10:09                  -3568.07860  0      11
        #                     log10-error:    Total        Iterations:
        #           Time      WFS    Density  Energy       Fermi  Poisson
        #iter:   0  12:13:29  +0.2            -3663.00195  0      1
        #iter:   1  12:15:18  -0.8            -4417.57264  0      1
        #iter:   2  12:17:07  -1.3            -4469.68829  0      1
        #iter:   3  12:18:57  -0.9   -0.8     -4091.42827  0      7
        #iter:   4  12:20:48  -1.1   -1.0     -4055.26110  0      7
        #iter:   5  12:22:40  -1.4   -1.3     -4106.38102  0      7
        #
        # new style
        #                     log10-error:    Total        Iterations:
        #           Time      WFS    Density  Energy       Fermi  Poisson
        #iter:   1  16:16:59  +0.4            -3663.00195  0      1
        #iter:   2  16:19:11  -0.8            -4417.57264  0      1
        #iter:   3  16:21:21  -1.3            -4469.68829  0      1
        #iter:   4  16:23:34  -0.9   -0.8     -4091.42827  0      7
        #iter:   5  16:25:49  -1.1   -1.0     -4055.26110  1      7
        #
        if len(str(gpaw_version)) == 1:
            # stable release found
            # assume new version - this is wrong but nothing else can be done
            start_iter = 1
        if len(str(gpaw_version)) > 4:
            # more then 4 digits in svnversion found
            # assume new version (can't compare strings here)
            start_iter = 1
        else:
            if gpaw_version >= 3172:
                # new version
                start_iter = 1
        # extract start time
        for n, l in enumerate(lines):
            if l.startswith('Date: '):
                #print l, n, f
                t = T(lines[n + 0].split()[4])
                break
        time[p]['start'] = t
        # extract SCF beginning time estimate and end time (constant potential steps (fixdensity))
        for n, l in enumerate(lines):
            if l.startswith('iter:   1'):
                #print l, n, f
                if start_iter == 0:
                    fixdensity_start = n-1
                else:
                    fixdensity_start = n+0
                t1 = T(lines[fixdensity_start + 0].split()[2])
                t2 = T(lines[fixdensity_start + 1].split()[2])
                t3 = T(lines[fixdensity_start + 2].split()[2])
                break
        # estimate the beginning of fixdensity based on 3 fixdensity steps
        time[p]['fixdensity_start_estimate'] = t1-(t3-t1)/2.0
        time[p]['fixdensity_end'] = t3
        # extract SCF beginning and end time
        time[p]['SCF_start'] = time[p]['fixdensity_end']
        for n, l in enumerate(lines):
            if l.startswith('iter: '+"%3s" % iter):
                #print l, n, f
                t = T(lines[n + 0].split()[2])
                break
        time[p]['SCF_end'] = t
        for n, l in enumerate(lines):
            if l.startswith('Total:') and (l.find('MB') == -1) and (l.find('GB') == -1):
                #print l, n, f
                t = l.split()[1]
                break
        time[p]['total'] = t
    #
    # results
    #
    speedup = {}
    efficiency = {}
    #
    if verbose:
        print("# p - processes, p0 - reference processes, t - time [sec], s - speedup, e - efficiency")
        print("# GPAW version "+str(gpaw_version)+": stages: 1 - initialization, 2 - fixdensity, 3 - SCF, 4 - forces, 5 - total")
        print("# p    "+" p/p0  "+" t1     "+" s1     "+" e1   "+" t2     "+" s2     "+" e2   "+" t3     "+" s3     "+" e3   "+" t4     "+" s4     "+" e4   "+" t5     "+" s5     "+" e5")
    for p in processes:
        time[p]['init'] = time[p]['fixdensity_start_estimate'] - time[p]['start']
        time[p]['fixdensity'] = time[p]['fixdensity_end'] - time[p]['fixdensity_start_estimate']
        time[p]['SCF'] = time[p]['SCF_end'] - time[p]['SCF_start']
        time[p]['forces'] = time[p]['forces_end'] - time[p]['forces_start']
        tot = max(time[p]['fixdensity_end'], time[p]['SCF_end'], time[p]['forces_end'])-time[p]['start']
        sum_of_entries = time[p]['init'] + time[p]['fixdensity']+ time[p]['SCF'] + time[p]['forces']
        #print time[p]['init'], time[p]['fixdensity'], time[p]['SCF'], time[p]['forces']
        if verbose:
            if abs(float(tot)-float(time[p]['total'])) > 5.0:
                print('Warning: Sum of time entries: '+str(tot)+' does not match total time in the output: '+str(time[p]['total']))
        time[p]['total'] = sum_of_entries
        # calculate
        speedup[p] = {}
        efficiency[p] = {}
        for stage in ['init', 'fixdensity', 'SCF', 'forces', 'total']:
            if time[p][stage] > 0.0:
                speedup[p][stage] = float(time[processes[0]][stage])/float(time[p][stage])*processes[0]
            else:
                speedup[p][stage] = 0.0
            efficiency[p][stage] = speedup[p][stage]/p
        # print results
        print('  %5d %6.2f %7.1f %7.1f %5.2f %7.1f %7.1f %5.2f %7.1f %7.1f %5.2f %7.1f %7.1f %5.2f %7.1f %7.1f %5.2f' %(
            p, float(p)/processes[0],
            float(time[p]['init']), speedup[p]['init'], efficiency[p]['init'],
            float(time[p]['fixdensity']), speedup[p]['fixdensity'], efficiency[p]['fixdensity'],
            float(time[p]['SCF']), speedup[p]['SCF'], efficiency[p]['SCF'],
            float(time[p]['forces']), speedup[p]['forces'], efficiency[p]['forces'],
            float(time[p]['total']), speedup[p]['total'], efficiency[p]['total'],
            ))

if __name__ == '__main__':
    from os import environ

    assert len(args) == 1, 'Error: Only one argument allowed: output prefix'
    output_prefix = args[0]

    analyse_benchmark(opt.dir, opt.pattern, output_prefix, opt.iter, opt.verbose)



scalapack = True

compiler = 'cc'
mpicompiler = compiler
mpilinker = mpicompiler

extra_compile_args += [
    '-O3',
    '-funroll-all-loops',
    '-fPIC',
    ]

libraries= []

dir_base = '/autofs/na1_home/farberow/sw/xt5/'
acml_base = '/opt/acml/4.1.0/gfortran64/lib/'
numpy_base = dir_base+'numpy-1.2.1/build/lib.linux-x86_64-2.5/numpy/'


extra_link_args = [
    '-L/usr/lib64 -lreadline -lncurses',
    numpy_base+'core/multiarray.a',
    numpy_base+'core/_sort.a',
    numpy_base+'core/scalarmath.a',
    numpy_base+'core/umath.a',
    numpy_base+'lib/_compiled_base.a',
    numpy_base+'numarray/_capi.a',
    numpy_base+'fft/fftpack_lite.a',
    numpy_base+'linalg/lapack_lite.a',
    numpy_base+'random/mtrand.a',
    '-L'+dir_base+'zlib-1.2.3-1/lib -lz',
    '-L'+dir_base+'expat-2.0.1-1/lib -lexpat',
    '-L'+acml_base+' -lacml',
]

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]
define_macros += [('NO_SOCKET', '1')]



import pytest
from gpaw import GPAW
from gpaw.point_groups import SymmetryChecker, PointGroup


@pytest.mark.serial
def test_c2v(gpw_files):
    calc = GPAW(gpw_files['h2o_lcao_wfs'])
    C = calc.atoms.positions[0]
    pg = PointGroup('C2v')
    sc = SymmetryChecker(pg, C, 2.0)
    symmetries = ''
    for n in range(4):
        print('-' * 70)
        dct = sc.check_band(calc, n)
        sym = dct['symmetry']
        symmetries += sym
        assert dct['characters'][sym] == pytest.approx(1, abs=0.01)
    assert symmetries == 'A1B2A1B1'



from myqueue.task import task


def create_tasks():
    return [
        task('bulk.py@4:6m'),
        task('surface.py@4:6m'),
        task('sigma.py', deps='bulk.py,surface.py'),
        task('fig2.py', deps='sigma.py')]



from gpaw.tddft import TDDFT
from gpaw.inducedfield.inducedfield_fdtd import (
    FDTDInducedField, calculate_hybrid_induced_field)
from gpaw.inducedfield.inducedfield_tddft import TDDFTInducedField

td_calc = TDDFT('td.gpw')

# Classical subsystem
cl_ind = FDTDInducedField(filename='cl.ind', paw=td_calc)
cl_ind.calculate_induced_field(gridrefinement=2)
cl_ind.write('cl_field.ind', mode='all')

# Quantum subsystem
qm_ind = TDDFTInducedField(filename='qm.ind', paw=td_calc)
qm_ind.calculate_induced_field(gridrefinement=2)
qm_ind.write('qm_field.ind', mode='all')

# Total system, interpolate/extrapolate to a grid with spacing h
tot_ind = calculate_hybrid_induced_field(cl_ind, qm_ind, h=1.0)
tot_ind.write('tot_field.ind', mode='all')



import numpy as np
from pathlib import Path
from gpaw.mpi import world
from gpaw.response.bse import BSE
from gpaw.response.df import DielectricFunction

ecut = 50
eshift = 0.8
eta = 0.2

df = DielectricFunction('gs_Si.gpw',
                        ecut=ecut,
                        frequencies=np.linspace(0, 10, 1001),
                        nbands=8,
                        intraband=False,
                        hilbert=False,
                        eta=eta,
                        eshift=eshift,
                        txt='rpa_Si.txt')

df.get_dielectric_function(filename='eps_rpa_Si.csv')

bse = BSE('gs_Si.gpw',
          ecut=ecut,
          valence_bands=range(0, 4),
          conduction_bands=range(4, 8),
          nbands=50,
          eshift=eshift,
          mode='BSE',
          write_v=True,
          integrate_gamma=0,
          txt='bse_Si.txt')

bse.get_dielectric_function(filename='eps_bse_Si.csv',
                            eta=eta,
                            write_eig='bse_Si_eig.dat',
                            w_w=np.linspace(0.0, 10.0, 10001))

if world.rank == 0:
    Path('v_TS.ulm').unlink()



from gpaw import GPAW
from ase.build import fcc111
from ase.build import add_adsorbate
atoms = fcc111('Pt', (2, 2, 6), a=4.00, vacuum=10.0)
add_adsorbate(atoms, 'O', 2.0, 'fcc')
calc = GPAW(mode='pw',
            kpts=(8, 8, 1),
            xc='RPBE')
ncpus = 4



from ase.build import molecule
from gpaw import GPAW
from gpaw.xc.libvdwxc import vdw_df_cx

atoms = molecule('H2O')
atoms.center(vacuum=3.0)
calc = GPAW(xc=vdw_df_cx())
atoms.calc = calc
atoms.get_potential_energy()



# Creates: Au.yml.png
import numpy as np
import matplotlib.pyplot as plt
from ase.units import _hplanck, _c, _e, Hartree

from gpaw.fdtd.polarizable_material import PermittivityPlus

_eps0_au = 1.0 / (4.0 * np.pi)


def eV_from_um(um_i):
    return _hplanck / _e * _c / (um_i * 1e-6)


def plot(fname, fiteps):
    with open(fname, 'r') as yml:
        for line in yml:
            if line.strip().startswith('data'):
                data_ij = np.array([[float(d) for d in line.split()]
                                    for line in yml])

    energy_j = eV_from_um(data_ij[:, 0])
    n_j = data_ij[:, 1]
    k_j = data_ij[:, 2]
    eps_j = (n_j ** 2 - k_j ** 2) + 1.0j * 2 * n_j * k_j

    energy_e = np.linspace(1.0, 6.0, 100)
    fiteps_e = np.array([fiteps.value(energy / Hartree) / _eps0_au
                         for energy in energy_e])

    plt.figure(figsize=(12, 6))
    plt.subplot(1, 2, 1)
    plt.plot(energy_j, eps_j.real, 'bv', label='data')
    plt.plot(energy_e, fiteps_e.real, 'b-', label='fit')
    plt.xlim(energy_e.min(), energy_e.max())
    # plt.ylim(fiteps_e.real.min(), fiteps_e.real.max())
    plt.ylim(-70, 0)
    plt.xlabel('Energy (eV)')
    plt.ylabel(r'Real($\epsilon$)')
    plt.legend(loc='best')

    plt.subplot(1, 2, 2)
    plt.plot(energy_j, eps_j.imag, 'bv')
    plt.plot(energy_e, fiteps_e.imag, 'b-')
    plt.xlim(energy_e.min(), energy_e.max())
    # plt.ylim(fiteps_e.imag.min(), fiteps_e.imag.max())
    plt.ylim(0, 7)
    plt.xlabel('Energy (eV)')
    plt.ylabel(r'Imaginary($\epsilon$)')
    plt.tight_layout()
    plt.savefig(f'{fname}.png')

# Permittivity of Gold
# Source:
# http://refractiveindex.info/?shelf=main&book=Au&page=Johnson
# Direct download link:
# wget http://refractiveindex.info/database/main/Au/Johnson.yml -O Au.yml


ymlfname = 'Au.yml'

# Fit to the permittivity
# J. Chem. Phys. 135, 084121 (2011); http://dx.doi.org/10.1063/1.3626549
fiteps = PermittivityPlus(data=[[0.2350, 0.1551, 95.62],
                                [0.4411, 0.1480, -12.55],
                                [0.7603, 1.946, -40.89],
                                [1.161, 1.396, 17.22],
                                [2.946, 1.183, 15.76],
                                [4.161, 1.964, 36.63],
                                [5.747, 1.958, 22.55],
                                [7.912, 1.361, 81.04]])
plot(ymlfname, fiteps)
